@@ -27,6 +27,7 @@
 
 #include "global.h"
 
+#include <assert.h>
 #include <ctype.h>
 #include <string.h>
 #include <pcre2.h>
@@ -51,7 +52,8 @@ static const UT_icd ut_int_pair_icd = {sizeof(int_pair),NULL,NULL,NULL};
 /* concatenate the string then move the pointer to the end */
 #define STRING_CAT(p, string, end)  do {    \
     size_t string_len = strlen(string); \
-    if (p + string_len >= end) \
+    assert(end > p); \
+    if (string_len >= (size_t)(end - p)) \
         return -1; \
     strcat(p, string); \
     p += string_len; \
@@ -192,10 +194,14 @@ int ec_glob(const char *pattern, const char *string)
                     if (!right_bracket)  /* The right bracket may not exist */
                         right_bracket = c + strlen(c);
 
-                    strcat(p_pcre, "\\");
+                    STRING_CAT(p_pcre, "\\", pcre_str_end);
+                    /* Boundary check for strncat below. */
+                    if (pcre_str_end - p_pcre <= right_bracket - c) {
+                        return -1;
+                    }
                     strncat(p_pcre, c, right_bracket - c);
                     if (*right_bracket)  /* right_bracket is a bracket */
-                        strcat(p_pcre, "\\]");
+                        STRING_CAT(p_pcre, "\\]", pcre_str_end);
                     p_pcre += strlen(p_pcre);
                     c = right_bracket;
                     if (!*c)
@@ -339,7 +345,7 @@ int ec_glob(const char *pattern, const char *string)
         }
     }
 
-    *(p_pcre ++) = '$';
+    ADD_CHAR(p_pcre, '$', pcre_str_end);
 
     pcre2_code_free(re); /* ^\\d+\\.\\.\\d+$ */
 
